/**********************

 QDP_D header file

**********************/

#ifndef _QDP_D
#define _QDP_D

#include "qdp_common.h"
#include "qdp_types.h"
#include "qla_types.h"
#include "qla_complex.h"
#include "qla_random.h"

#ifdef __cplusplus
extern "C" {
#endif

/* create and destroy fields */

QDP_D_Real * QDP_D_create_R(void);
void QDP_D_destroy_R(QDP_D_Real *field);
QDP_D_Complex * QDP_D_create_C(void);
void QDP_D_destroy_C(QDP_D_Complex *field);

/* data access and extraction */

QLA_D_Real * QDP_D_expose_R(QDP_D_Real *field);
void QDP_D_reset_R(QDP_D_Real *field);
void QDP_D_extract_R(QLA_D_Real *dest, QDP_D_Real *src, QDP_Subset subset);
void QDP_D_insert_R(QDP_D_Real *dest, QLA_D_Real *src, QDP_Subset subset);
QLA_D_Complex * QDP_D_expose_C(QDP_D_Complex *field);
void QDP_D_reset_C(QDP_D_Complex *field);
void QDP_D_extract_C(QLA_D_Complex *dest, QDP_D_Complex *src, QDP_Subset subset);
void QDP_D_insert_C(QDP_D_Complex *dest, QLA_D_Complex *src, QDP_Subset subset);

/* optimization */

void QDP_D_discard_R(QDP_D_Real *field);
void QDP_D_discard_C(QDP_D_Complex *field);

/* set by function */

void QDP_D_R_eq_func(QDP_D_Real *dest, void (*func)(QLA_D_Real *dest, int coords[]), QDP_Subset subset );
void QDP_D_C_eq_func(QDP_D_Complex *dest, void (*func)(QLA_D_Complex *dest, int coords[]), QDP_Subset subset );

/* shift */

void QDP_D_R_eq_sR(QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Shift shift, QDP_ShiftDir fb, QDP_Subset subset);
void QDP_D_R_veq_sR(QDP_D_Real *__restrict__ dest[], QDP_D_Real *src[], QDP_Shift shift[], QDP_ShiftDir fb[], QDP_Subset subset, int nv);
void QDP_D_C_eq_sC(QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Shift shift, QDP_ShiftDir fb, QDP_Subset subset);
void QDP_D_C_veq_sC(QDP_D_Complex *__restrict__ dest[], QDP_D_Complex *src[], QDP_Shift shift[], QDP_ShiftDir fb[], QDP_Subset subset, int nv);

/* shift & multiply */


/*
 *  Fills and random numbers
 */

/*
 *  Zero fills
 */

void QDP_D_R_eq_zero( QDP_D_Real *__restrict__ dest, QDP_Subset subset );
void QDP_D_C_eq_zero( QDP_D_Complex *__restrict__ dest, QDP_Subset subset );

/*
 *  Constant fills
 */

void QDP_D_R_eq_r( QDP_D_Real *__restrict__ dest, QLA_D_Real *src, QDP_Subset subset );
void QDP_D_C_eq_c( QDP_D_Complex *__restrict__ dest, QLA_D_Complex *src, QDP_Subset subset );

/*
 *  Uniform random number fills
 */

void QDP_D_R_eq_random_S( QDP_D_Real *__restrict__ dest, QDP_RandomState *src, QDP_Subset subset );

/*
 *  Gaussian random number fills
 */

void QDP_D_R_eq_gaussian_S( QDP_D_Real *__restrict__ dest, QDP_RandomState *src, QDP_Subset subset );
void QDP_D_C_eq_gaussian_S( QDP_D_Complex *__restrict__ dest, QDP_RandomState *src, QDP_Subset subset );

/*
 *  Unary Operations
 */

/*
 *  Elementary unary functions on reals
 */

void QDP_D_R_eq_sin_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_cos_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_tan_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_asin_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_acos_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_atan_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_sqrt_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_fabs_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_exp_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_log_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_sign_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_ceil_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_floor_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_sinh_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_cosh_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_tanh_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_R_eq_log10_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );

/*
 *  Elementary unary functions real to complex
 */

void QDP_D_C_eq_cexpi_R( QDP_D_Complex *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );

/*
 *  Elementary unary functions complex to real
 */

void QDP_D_R_eq_norm_C( QDP_D_Real *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_R_eq_arg_C( QDP_D_Real *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );

/*
 *  Elementary unary functions on complex values
 */

void QDP_D_C_eq_cexp_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_C_eq_csqrt_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_C_eq_clog_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );

/*
 *  Copying
 */

void QDP_D_R_eq_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_C_eq_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_R_veq_R( QDP_D_Real *__restrict__ dest[], QDP_D_Real *src[], QDP_Subset subset, int nv );
void QDP_D_C_veq_C( QDP_D_Complex *__restrict__ dest[], QDP_D_Complex *src[], QDP_Subset subset, int nv );

/*
 *  Incrementing
 */

void QDP_D_R_eqm_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_C_eqm_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_R_peq_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_C_peq_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_R_meq_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_C_meq_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_R_veqm_R( QDP_D_Real *__restrict__ dest[], QDP_D_Real *src[], QDP_Subset subset, int nv );
void QDP_D_C_veqm_C( QDP_D_Complex *__restrict__ dest[], QDP_D_Complex *src[], QDP_Subset subset, int nv );
void QDP_D_R_vpeq_R( QDP_D_Real *__restrict__ dest[], QDP_D_Real *src[], QDP_Subset subset, int nv );
void QDP_D_C_vpeq_C( QDP_D_Complex *__restrict__ dest[], QDP_D_Complex *src[], QDP_Subset subset, int nv );
void QDP_D_R_vmeq_R( QDP_D_Real *__restrict__ dest[], QDP_D_Real *src[], QDP_Subset subset, int nv );
void QDP_D_C_vmeq_C( QDP_D_Complex *__restrict__ dest[], QDP_D_Complex *src[], QDP_Subset subset, int nv );

/*
 *  Complex conjugate
 */

void QDP_D_C_eq_conj_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_C_peq_conj_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_C_meq_conj_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_C_eqm_conj_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );

/*
 *  Hermitian conjugate
 */

void QDP_D_C_eq_Ca( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_C_peq_Ca( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_C_meq_Ca( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_C_eqm_Ca( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );

/*
 *  Local squared norm: uniform precision
 */

void QDP_D_R_eq_norm2_C( QDP_D_Real *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );

/*
 *  Type conversion and component extraction and insertion
 */

/*
 *  Convert real to complex (zero imaginary part)
 */

void QDP_D_C_eq_R( QDP_D_Complex *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );

/*
 *  Convert real and imaginary to complex
 */

void QDP_D_C_eq_R_plus_i_R( QDP_D_Complex *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );

/*
 *  Real/Imaginary part of complex
 */

void QDP_D_R_eq_re_C( QDP_D_Real *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_R_eq_im_C( QDP_D_Real *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );

/*
 *  Integer to real
 */

void QDP_D_R_eq_I( QDP_D_Real *__restrict__ dest, QDP_Int *src, QDP_Subset subset );

/*
 *  Real to integer (truncate/round)
 */

void QDP_D_I_eq_trunc_R( QDP_Int *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_I_eq_round_R( QDP_Int *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );

/*
 *  Binary operations with constants
 */

/*
 *  Multiplication by real constant
 */

void QDP_D_R_eq_r_times_R( QDP_D_Real *__restrict__ dest, QLA_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_C_eq_r_times_C( QDP_D_Complex *__restrict__ dest, QLA_D_Real *src1, QDP_D_Complex *src2, QDP_Subset subset );
void QDP_D_R_peq_r_times_R( QDP_D_Real *__restrict__ dest, QLA_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_C_peq_r_times_C( QDP_D_Complex *__restrict__ dest, QLA_D_Real *src1, QDP_D_Complex *src2, QDP_Subset subset );
void QDP_D_R_meq_r_times_R( QDP_D_Real *__restrict__ dest, QLA_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_C_meq_r_times_C( QDP_D_Complex *__restrict__ dest, QLA_D_Real *src1, QDP_D_Complex *src2, QDP_Subset subset );
void QDP_D_R_eqm_r_times_R( QDP_D_Real *__restrict__ dest, QLA_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_C_eqm_r_times_C( QDP_D_Complex *__restrict__ dest, QLA_D_Real *src1, QDP_D_Complex *src2, QDP_Subset subset );

/*
 *  Multiplication by complex constant
 */

void QDP_D_C_eq_c_times_C( QDP_D_Complex *__restrict__ dest, QLA_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );
void QDP_D_C_peq_c_times_C( QDP_D_Complex *__restrict__ dest, QLA_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );
void QDP_D_C_meq_c_times_C( QDP_D_Complex *__restrict__ dest, QLA_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );
void QDP_D_C_eqm_c_times_C( QDP_D_Complex *__restrict__ dest, QLA_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );

/*
 *  Multiplication by i
 */

void QDP_D_C_eq_i_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_C_peq_i_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_C_meq_i_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_C_eqm_i_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );

/*
 *  Binary operations with fields
 */

/*
 *  Elementary binary functions on reals
 */

void QDP_D_R_eq_R_mod_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_R_eq_R_max_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_R_eq_R_min_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_R_eq_R_pow_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_R_eq_R_atan2_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );

/*
 *  Multiplying real by integer power of 2
 */

void QDP_D_R_eq_R_ldexp_I( QDP_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_Int *src2, QDP_Subset subset );

/*
 *  Addition
 */

void QDP_D_R_eq_R_plus_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_C_eq_C_plus_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );

/*
 *  Subtraction
 */

void QDP_D_R_eq_R_minus_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_C_eq_C_minus_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );

/*
 *  Multiplication: uniform types
 */

void QDP_D_R_eq_R_times_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_C_eq_C_times_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );
void QDP_D_R_peq_R_times_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_C_peq_C_times_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );
void QDP_D_R_meq_R_times_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_C_meq_C_times_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );
void QDP_D_R_eqm_R_times_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_C_eqm_C_times_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );

/*
 *  Division of integer, real, and complex fields
 */

void QDP_D_R_eq_R_divide_R( QDP_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_C_eq_C_divide_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );

/*
 *  Local inner product
 */

void QDP_D_C_eq_C_dot_C( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );

/*
 *  Real part of local inner product
 */

void QDP_D_R_eq_re_C_dot_C( QDP_D_Real *__restrict__ dest, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );

/*
 *  Ternary operations with fields
 */

/*
 *  Addition or subtraction with real scalar multiplication
 */

void QDP_D_R_eq_r_times_R_plus_R( QDP_D_Real *__restrict__ dest, QLA_D_Real *c, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_C_eq_r_times_C_plus_C( QDP_D_Complex *__restrict__ dest, QLA_D_Real *c, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );
void QDP_D_R_eq_r_times_R_minus_R( QDP_D_Real *__restrict__ dest, QLA_D_Real *c, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_C_eq_r_times_C_minus_C( QDP_D_Complex *__restrict__ dest, QLA_D_Real *c, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );

/*
 *  Addition or subtraction with complex scalar multiplication
 */

void QDP_D_C_eq_c_times_C_plus_C( QDP_D_Complex *__restrict__ dest, QLA_D_Complex *c, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );
void QDP_D_C_eq_c_times_C_minus_C( QDP_D_Complex *__restrict__ dest, QLA_D_Complex *c, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );

/*
 *  Boolean operations
 */

/*
 *  Comparisons of integers and reals
 */

void QDP_D_I_eq_R_eq_R( QDP_Int *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_I_eq_R_ne_R( QDP_Int *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_I_eq_R_gt_R( QDP_Int *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_I_eq_R_lt_R( QDP_Int *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_I_eq_R_ge_R( QDP_Int *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_I_eq_R_le_R( QDP_Int *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );

/*
 *  Copy under a mask
 */

void QDP_D_R_eq_R_mask_I( QDP_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_D_C_eq_C_mask_I( QDP_D_Complex *__restrict__ dest, QDP_D_Complex *src1, QDP_Int *src2, QDP_Subset subset );

/*
 *  Reductions
 */

/*
 *  Global squared norm: uniform precision
 */

void QDP_D_r_eq_norm2_I( QLA_D_Real *__restrict__ dest, QDP_Int *src, QDP_Subset subset );
void QDP_D_r_eq_norm2_R( QLA_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_r_eq_norm2_C( QLA_D_Real *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );
void QDP_D_r_veq_norm2_I( QLA_D_Real dest[], QDP_Int *src[], QDP_Subset subset, int nv );
void QDP_D_r_veq_norm2_R( QLA_D_Real dest[], QDP_D_Real *src[], QDP_Subset subset, int nv );
void QDP_D_r_veq_norm2_C( QLA_D_Real dest[], QDP_D_Complex *src[], QDP_Subset subset, int nv );

/*
 *  Global inner product
 */

void QDP_D_r_eq_I_dot_I( QLA_D_Real *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_D_r_eq_R_dot_R( QLA_D_Real *__restrict__ dest, QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset );
void QDP_D_r_veq_I_dot_I( QLA_D_Real dest[], QDP_Int *src1[], QDP_Int *src2[], QDP_Subset subset, int nv );
void QDP_D_r_veq_R_dot_R( QLA_D_Real dest[], QDP_D_Real *src1[], QDP_D_Real *src2[], QDP_Subset subset, int nv );
void QDP_D_c_eq_C_dot_C( QLA_D_Complex *__restrict__ dest, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );
void QDP_D_c_veq_C_dot_C( QLA_D_Complex dest[], QDP_D_Complex *src1[], QDP_D_Complex *src2[], QDP_Subset subset, int nv );

/*
 *  Real part of global inner product
 */

void QDP_D_r_eq_re_C_dot_C( QLA_D_Real *__restrict__ dest, QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset );
void QDP_D_r_veq_re_C_dot_C( QLA_D_Real dest[], QDP_D_Complex *src1[], QDP_D_Complex *src2[], QDP_Subset subset, int nv );

/*
 *  Global sums
 */

void QDP_D_r_eq_sum_I( QLA_D_Real *__restrict__ dest, QDP_Int *src, QDP_Subset subset );
void QDP_D_r_eq_sum_R( QLA_D_Real *__restrict__ dest, QDP_D_Real *src, QDP_Subset subset );
void QDP_D_c_eq_sum_C( QLA_D_Complex *__restrict__ dest, QDP_D_Complex *src, QDP_Subset subset );

/*
 *  Multisubset Norms
 */

void QDP_D_r_eq_norm2_I_multi( QLA_D_Real dest[], QDP_Int *src, QDP_Subset subset[], int ns );
void QDP_D_r_eq_norm2_R_multi( QLA_D_Real dest[], QDP_D_Real *src, QDP_Subset subset[], int ns );
void QDP_D_r_eq_norm2_C_multi( QLA_D_Real dest[], QDP_D_Complex *src, QDP_Subset subset[], int ns );
void QDP_D_r_veq_norm2_I_multi( QLA_D_Real dest[], QDP_Int *src[], QDP_Subset subset[], int ns );
void QDP_D_r_veq_norm2_R_multi( QLA_D_Real dest[], QDP_D_Real *src[], QDP_Subset subset[], int ns );
void QDP_D_r_veq_norm2_C_multi( QLA_D_Real dest[], QDP_D_Complex *src[], QDP_Subset subset[], int ns );

/*
 *  Multisubset inner products
 */

void QDP_D_r_eq_I_dot_I_multi( QLA_D_Real dest[], QDP_Int *src1, QDP_Int *src2, QDP_Subset subset[], int ns );
void QDP_D_r_eq_R_dot_R_multi( QLA_D_Real dest[], QDP_D_Real *src1, QDP_D_Real *src2, QDP_Subset subset[], int ns );
void QDP_D_r_veq_I_dot_I_multi( QLA_D_Real dest[], QDP_Int *src1[], QDP_Int *src2[], QDP_Subset subset[], int ns );
void QDP_D_r_veq_R_dot_R_multi( QLA_D_Real dest[], QDP_D_Real *src1[], QDP_D_Real *src2[], QDP_Subset subset[], int ns );
void QDP_D_c_eq_C_dot_C_multi( QLA_D_Complex dest[], QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset[], int ns );
void QDP_D_c_veq_C_dot_C_multi( QLA_D_Complex dest[], QDP_D_Complex *src1[], QDP_D_Complex *src2[], QDP_Subset subset[], int ns );

/*
 *  Multisubset real part of global inner product
 */

void QDP_D_r_eq_re_C_dot_C_multi( QLA_D_Real dest[], QDP_D_Complex *src1, QDP_D_Complex *src2, QDP_Subset subset[], int ns );
void QDP_D_r_veq_re_C_dot_C_multi( QLA_D_Real dest[], QDP_D_Complex *src1[], QDP_D_Complex *src2[], QDP_Subset subset[], int ns );

/*
 *  Multisubset global sums
 */

void QDP_D_r_eq_sum_I_multi( QLA_D_Real dest[], QDP_Int *src, QDP_Subset subset[], int ns );
void QDP_D_r_eq_sum_R_multi( QLA_D_Real dest[], QDP_D_Real *src, QDP_Subset subset[], int ns );
void QDP_D_c_eq_sum_C_multi( QLA_D_Complex dest[], QDP_D_Complex *src, QDP_Subset subset[], int ns );

#ifdef __cplusplus
}
#endif

#ifdef QDP_PROFILE
#include "qdp_d_profile.h"
#endif

#endif
