/**********************

 QDP_F3 header file

**********************/

#ifndef _QDP_F3
#define _QDP_F3

#include "qdp_common.h"
#include "qdp_types.h"
#include "qla_types.h"
#include "qla_complex.h"
#include "qla_random.h"

#ifdef __cplusplus
extern "C" {
#endif

/* create and destroy fields */

QDP_F3_ColorVector * QDP_F3_create_V(void);
void QDP_F3_destroy_V(QDP_F3_ColorVector *field);
QDP_F3_HalfFermion * QDP_F3_create_H(void);
void QDP_F3_destroy_H(QDP_F3_HalfFermion *field);
QDP_F3_DiracFermion * QDP_F3_create_D(void);
void QDP_F3_destroy_D(QDP_F3_DiracFermion *field);
QDP_F3_ColorMatrix * QDP_F3_create_M(void);
void QDP_F3_destroy_M(QDP_F3_ColorMatrix *field);
QDP_F3_DiracPropagator * QDP_F3_create_P(void);
void QDP_F3_destroy_P(QDP_F3_DiracPropagator *field);

/* data access and extraction */

QLA_F3_ColorVector * QDP_F3_expose_V(QDP_F3_ColorVector *field);
void QDP_F3_reset_V(QDP_F3_ColorVector *field);
void QDP_F3_extract_V(QLA_F3_ColorVector *dest, QDP_F3_ColorVector *src, QDP_Subset subset);
void QDP_F3_insert_V(QDP_F3_ColorVector *dest, QLA_F3_ColorVector *src, QDP_Subset subset);
QLA_F3_HalfFermion * QDP_F3_expose_H(QDP_F3_HalfFermion *field);
void QDP_F3_reset_H(QDP_F3_HalfFermion *field);
void QDP_F3_extract_H(QLA_F3_HalfFermion *dest, QDP_F3_HalfFermion *src, QDP_Subset subset);
void QDP_F3_insert_H(QDP_F3_HalfFermion *dest, QLA_F3_HalfFermion *src, QDP_Subset subset);
QLA_F3_DiracFermion * QDP_F3_expose_D(QDP_F3_DiracFermion *field);
void QDP_F3_reset_D(QDP_F3_DiracFermion *field);
void QDP_F3_extract_D(QLA_F3_DiracFermion *dest, QDP_F3_DiracFermion *src, QDP_Subset subset);
void QDP_F3_insert_D(QDP_F3_DiracFermion *dest, QLA_F3_DiracFermion *src, QDP_Subset subset);
QLA_F3_ColorMatrix * QDP_F3_expose_M(QDP_F3_ColorMatrix *field);
void QDP_F3_reset_M(QDP_F3_ColorMatrix *field);
void QDP_F3_extract_M(QLA_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src, QDP_Subset subset);
void QDP_F3_insert_M(QDP_F3_ColorMatrix *dest, QLA_F3_ColorMatrix *src, QDP_Subset subset);
QLA_F3_DiracPropagator * QDP_F3_expose_P(QDP_F3_DiracPropagator *field);
void QDP_F3_reset_P(QDP_F3_DiracPropagator *field);
void QDP_F3_extract_P(QLA_F3_DiracPropagator *dest, QDP_F3_DiracPropagator *src, QDP_Subset subset);
void QDP_F3_insert_P(QDP_F3_DiracPropagator *dest, QLA_F3_DiracPropagator *src, QDP_Subset subset);

/* optimization */

void QDP_F3_discard_V(QDP_F3_ColorVector *field);
void QDP_F3_discard_H(QDP_F3_HalfFermion *field);
void QDP_F3_discard_D(QDP_F3_DiracFermion *field);
void QDP_F3_discard_M(QDP_F3_ColorMatrix *field);
void QDP_F3_discard_P(QDP_F3_DiracPropagator *field);

/* set by function */

void QDP_F3_V_eq_func(QDP_F3_ColorVector *dest, void (*func)(QLA_F3_ColorVector *dest, int coords[]), QDP_Subset subset );
void QDP_F3_H_eq_func(QDP_F3_HalfFermion *dest, void (*func)(QLA_F3_HalfFermion *dest, int coords[]), QDP_Subset subset );
void QDP_F3_D_eq_func(QDP_F3_DiracFermion *dest, void (*func)(QLA_F3_DiracFermion *dest, int coords[]), QDP_Subset subset );
void QDP_F3_M_eq_func(QDP_F3_ColorMatrix *dest, void (*func)(QLA_F3_ColorMatrix *dest, int coords[]), QDP_Subset subset );
void QDP_F3_P_eq_func(QDP_F3_DiracPropagator *dest, void (*func)(QLA_F3_DiracPropagator *dest, int coords[]), QDP_Subset subset );

/* shift */

void QDP_F3_V_eq_sV(QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Shift shift, QDP_ShiftDir fb, QDP_Subset subset);
void QDP_F3_V_veq_sV(QDP_F3_ColorVector *__restrict__ dest[], QDP_F3_ColorVector *src[], QDP_Shift shift[], QDP_ShiftDir fb[], QDP_Subset subset, int nv);
void QDP_F3_H_eq_sH(QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Shift shift, QDP_ShiftDir fb, QDP_Subset subset);
void QDP_F3_H_veq_sH(QDP_F3_HalfFermion *__restrict__ dest[], QDP_F3_HalfFermion *src[], QDP_Shift shift[], QDP_ShiftDir fb[], QDP_Subset subset, int nv);
void QDP_F3_D_eq_sD(QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Shift shift, QDP_ShiftDir fb, QDP_Subset subset);
void QDP_F3_D_veq_sD(QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_DiracFermion *src[], QDP_Shift shift[], QDP_ShiftDir fb[], QDP_Subset subset, int nv);
void QDP_F3_M_eq_sM(QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Shift shift, QDP_ShiftDir fb, QDP_Subset subset);
void QDP_F3_M_veq_sM(QDP_F3_ColorMatrix *__restrict__ dest[], QDP_F3_ColorMatrix *src[], QDP_Shift shift[], QDP_ShiftDir fb[], QDP_Subset subset, int nv);
void QDP_F3_P_eq_sP(QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Shift shift, QDP_ShiftDir fb, QDP_Subset subset);
void QDP_F3_P_veq_sP(QDP_F3_DiracPropagator *__restrict__ dest[], QDP_F3_DiracPropagator *src[], QDP_Shift shift[], QDP_ShiftDir fb[], QDP_Subset subset, int nv);

/* shift & multiply */

void QDP_F3_V_eq_M_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_eq_Ma_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_eq_sM_times_V(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_eq_sMa_times_V(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_eq_sM_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_eq_sMa_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_peq_M_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_peq_Ma_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_peq_sM_times_V(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_peq_sMa_times_V(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_peq_sM_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_peq_sMa_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_meq_M_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_meq_Ma_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_meq_sM_times_V(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_meq_sMa_times_V(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_meq_sM_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_meq_sMa_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_eqm_M_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_eqm_Ma_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_eqm_sM_times_V(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_eqm_sMa_times_V(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_eqm_sM_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_V_eqm_sMa_times_sV(QDP_F3_ColorVector *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_eq_M_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_eq_Ma_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_eq_sM_times_H(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_eq_sMa_times_H(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_eq_sM_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_eq_sMa_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_peq_M_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_peq_Ma_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_peq_sM_times_H(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_peq_sMa_times_H(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_peq_sM_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_peq_sMa_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_meq_M_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_meq_Ma_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_meq_sM_times_H(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_meq_sMa_times_H(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_meq_sM_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_meq_sMa_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_eqm_M_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_eqm_Ma_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_eqm_sM_times_H(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_eqm_sMa_times_H(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_eqm_sM_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_H_eqm_sMa_times_sH(QDP_F3_HalfFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_eq_M_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_eq_Ma_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_eq_sM_times_D(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_eq_sMa_times_D(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_eq_sM_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_eq_sMa_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_peq_M_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_peq_Ma_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_peq_sM_times_D(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_peq_sMa_times_D(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_peq_sM_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_peq_sMa_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_meq_M_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_meq_Ma_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_meq_sM_times_D(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_meq_sMa_times_D(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_meq_sM_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_meq_sMa_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_eqm_M_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_eqm_Ma_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_eqm_sM_times_D(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_eqm_sMa_times_D(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_eqm_sM_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_D_eqm_sMa_times_sD(QDP_F3_DiracFermion *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_eq_M_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_eq_Ma_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_eq_sM_times_M(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_eq_sMa_times_M(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_eq_sM_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_eq_sMa_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_peq_M_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_peq_Ma_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_peq_sM_times_M(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_peq_sMa_times_M(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_peq_sM_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_peq_sMa_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_meq_M_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_meq_Ma_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_meq_sM_times_M(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_meq_sMa_times_M(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_meq_sM_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_meq_sMa_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_eqm_M_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_eqm_Ma_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_eqm_sM_times_M(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_eqm_sMa_times_M(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_eqm_sM_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_M_eqm_sMa_times_sM(QDP_F3_ColorMatrix *dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_eq_M_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_eq_Ma_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_eq_sM_times_P(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_eq_sMa_times_P(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_eq_sM_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_eq_sMa_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_peq_M_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_peq_Ma_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_peq_sM_times_P(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_peq_sMa_times_P(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_peq_sM_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_peq_sMa_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_meq_M_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_meq_Ma_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_meq_sM_times_P(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_meq_sMa_times_P(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_meq_sM_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_meq_sMa_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_eqm_M_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_eqm_Ma_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_eqm_sM_times_P(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_eqm_sMa_times_P(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_eqm_sM_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);
void QDP_F3_P_eqm_sMa_times_sP(QDP_F3_DiracPropagator *dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Shift shift, int fb, QDP_Subset subset);

/*
 *  Fills and random numbers
 */

/*
 *  Zero fills
 */

void QDP_F3_V_eq_zero( QDP_F3_ColorVector *__restrict__ dest, QDP_Subset subset );
void QDP_F3_H_eq_zero( QDP_F3_HalfFermion *__restrict__ dest, QDP_Subset subset );
void QDP_F3_D_eq_zero( QDP_F3_DiracFermion *__restrict__ dest, QDP_Subset subset );
void QDP_F3_M_eq_zero( QDP_F3_ColorMatrix *__restrict__ dest, QDP_Subset subset );
void QDP_F3_P_eq_zero( QDP_F3_DiracPropagator *__restrict__ dest, QDP_Subset subset );

/*
 *  Constant fills
 */

void QDP_F3_V_eq_v( QDP_F3_ColorVector *__restrict__ dest, QLA_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_H_eq_h( QDP_F3_HalfFermion *__restrict__ dest, QLA_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_D_eq_d( QDP_F3_DiracFermion *__restrict__ dest, QLA_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_M_eq_m( QDP_F3_ColorMatrix *__restrict__ dest, QLA_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_eq_p( QDP_F3_DiracPropagator *__restrict__ dest, QLA_F3_DiracPropagator *src, QDP_Subset subset );

/*
 *  Fill color matrix with constant times identity
 */

void QDP_F3_M_eq_c( QDP_F3_ColorMatrix *__restrict__ dest, QLA_F_Complex *src, QDP_Subset subset );

/*
 *  Gaussian random number fills
 */

void QDP_F3_V_eq_gaussian_S( QDP_F3_ColorVector *__restrict__ dest, QDP_RandomState *src, QDP_Subset subset );
void QDP_F3_H_eq_gaussian_S( QDP_F3_HalfFermion *__restrict__ dest, QDP_RandomState *src, QDP_Subset subset );
void QDP_F3_D_eq_gaussian_S( QDP_F3_DiracFermion *__restrict__ dest, QDP_RandomState *src, QDP_Subset subset );
void QDP_F3_M_eq_gaussian_S( QDP_F3_ColorMatrix *__restrict__ dest, QDP_RandomState *src, QDP_Subset subset );
void QDP_F3_P_eq_gaussian_S( QDP_F3_DiracPropagator *__restrict__ dest, QDP_RandomState *src, QDP_Subset subset );

/*
 *  Unary Operations
 */

/*
 *  Copying
 */

void QDP_F3_V_eq_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_H_eq_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_D_eq_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_M_eq_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_eq_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_V_veq_V( QDP_F3_ColorVector *__restrict__ dest[], QDP_F3_ColorVector *src[], QDP_Subset subset, int nv );
void QDP_F3_H_veq_H( QDP_F3_HalfFermion *__restrict__ dest[], QDP_F3_HalfFermion *src[], QDP_Subset subset, int nv );
void QDP_F3_D_veq_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_DiracFermion *src[], QDP_Subset subset, int nv );
void QDP_F3_M_veq_M( QDP_F3_ColorMatrix *__restrict__ dest[], QDP_F3_ColorMatrix *src[], QDP_Subset subset, int nv );
void QDP_F3_P_veq_P( QDP_F3_DiracPropagator *__restrict__ dest[], QDP_F3_DiracPropagator *src[], QDP_Subset subset, int nv );

/*
 *  Incrementing
 */

void QDP_F3_V_eqm_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_H_eqm_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_D_eqm_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_M_eqm_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_eqm_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_V_peq_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_H_peq_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_D_peq_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_M_peq_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_peq_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_V_meq_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_H_meq_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_D_meq_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_M_meq_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_meq_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_V_veqm_V( QDP_F3_ColorVector *__restrict__ dest[], QDP_F3_ColorVector *src[], QDP_Subset subset, int nv );
void QDP_F3_H_veqm_H( QDP_F3_HalfFermion *__restrict__ dest[], QDP_F3_HalfFermion *src[], QDP_Subset subset, int nv );
void QDP_F3_D_veqm_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_DiracFermion *src[], QDP_Subset subset, int nv );
void QDP_F3_M_veqm_M( QDP_F3_ColorMatrix *__restrict__ dest[], QDP_F3_ColorMatrix *src[], QDP_Subset subset, int nv );
void QDP_F3_P_veqm_P( QDP_F3_DiracPropagator *__restrict__ dest[], QDP_F3_DiracPropagator *src[], QDP_Subset subset, int nv );
void QDP_F3_V_vpeq_V( QDP_F3_ColorVector *__restrict__ dest[], QDP_F3_ColorVector *src[], QDP_Subset subset, int nv );
void QDP_F3_H_vpeq_H( QDP_F3_HalfFermion *__restrict__ dest[], QDP_F3_HalfFermion *src[], QDP_Subset subset, int nv );
void QDP_F3_D_vpeq_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_DiracFermion *src[], QDP_Subset subset, int nv );
void QDP_F3_M_vpeq_M( QDP_F3_ColorMatrix *__restrict__ dest[], QDP_F3_ColorMatrix *src[], QDP_Subset subset, int nv );
void QDP_F3_P_vpeq_P( QDP_F3_DiracPropagator *__restrict__ dest[], QDP_F3_DiracPropagator *src[], QDP_Subset subset, int nv );
void QDP_F3_V_vmeq_V( QDP_F3_ColorVector *__restrict__ dest[], QDP_F3_ColorVector *src[], QDP_Subset subset, int nv );
void QDP_F3_H_vmeq_H( QDP_F3_HalfFermion *__restrict__ dest[], QDP_F3_HalfFermion *src[], QDP_Subset subset, int nv );
void QDP_F3_D_vmeq_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_DiracFermion *src[], QDP_Subset subset, int nv );
void QDP_F3_M_vmeq_M( QDP_F3_ColorMatrix *__restrict__ dest[], QDP_F3_ColorMatrix *src[], QDP_Subset subset, int nv );
void QDP_F3_P_vmeq_P( QDP_F3_DiracPropagator *__restrict__ dest[], QDP_F3_DiracPropagator *src[], QDP_Subset subset, int nv );

/*
 *  Transpose
 */

void QDP_F3_M_eq_transpose_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_eq_transpose_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_M_peq_transpose_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_peq_transpose_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_M_meq_transpose_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_meq_transpose_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_M_eqm_transpose_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_eqm_transpose_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );

/*
 *  Complex conjugate
 */

void QDP_F3_V_eq_conj_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_H_eq_conj_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_D_eq_conj_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_M_eq_conj_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_eq_conj_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_V_peq_conj_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_H_peq_conj_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_D_peq_conj_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_M_peq_conj_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_peq_conj_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_V_meq_conj_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_H_meq_conj_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_D_meq_conj_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_M_meq_conj_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_meq_conj_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_V_eqm_conj_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_H_eqm_conj_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_D_eqm_conj_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_M_eqm_conj_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_eqm_conj_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );

/*
 *  Hermitian conjugate
 */

void QDP_F3_M_eq_Ma( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_eq_Pa( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_M_peq_Ma( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_peq_Pa( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_M_meq_Ma( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_meq_Pa( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_M_eqm_Ma( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_eqm_Pa( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );

/*
 *  Local squared norm: uniform precision
 */

void QDP_F3_R_eq_norm2_V( QDP_F_Real *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_R_eq_norm2_H( QDP_F_Real *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_R_eq_norm2_D( QDP_F_Real *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_R_eq_norm2_M( QDP_F_Real *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_R_eq_norm2_P( QDP_F_Real *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );

/*
 *  Type conversion and component extraction and insertion
 */

/*
 *  Accessing a color matrix element
 */

void QDP_F3_C_eq_elem_M( QDP_F_Complex *__restrict__ dest, QDP_F3_ColorMatrix *src, int i, int j, QDP_Subset subset );

/*
 *  Inserting a color matrix element
 */

void QDP_F3_M_eq_elem_C( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F_Complex *src, int i, int j, QDP_Subset subset );

/*
 *  Accessing a half fermion or Dirac fermion spinor element
 */

void QDP_F3_C_eq_elem_H( QDP_F_Complex *__restrict__ dest, QDP_F3_HalfFermion *src, int color, int spin, QDP_Subset subset );
void QDP_F3_C_eq_elem_D( QDP_F_Complex *__restrict__ dest, QDP_F3_DiracFermion *src, int color, int spin, QDP_Subset subset );

/*
 *  Inserting a half fermion or Dirac fermion spinor element
 */

void QDP_F3_H_eq_elem_C( QDP_F3_HalfFermion *__restrict__ dest, QDP_F_Complex *src, int color, int spin, QDP_Subset subset );
void QDP_F3_D_eq_elem_C( QDP_F3_DiracFermion *__restrict__ dest, QDP_F_Complex *src, int color, int spin, QDP_Subset subset );

/*
 *  Accessing a color vector element
 */

void QDP_F3_C_eq_elem_V( QDP_F_Complex *__restrict__ dest, QDP_F3_ColorVector *src, int i, QDP_Subset subset );

/*
 *  Inserting a color vector element
 */

void QDP_F3_V_eq_elem_C( QDP_F3_ColorVector *__restrict__ dest, QDP_F_Complex *src, int i, QDP_Subset subset );

/*
 *  Accessing a Dirac propagator matrix element
 */

void QDP_F3_C_eq_elem_P( QDP_F_Complex *__restrict__ dest, QDP_F3_DiracPropagator *src, int ic, int is, int jc, int js, QDP_Subset subset );

/*
 *  Inserting a Dirac propagator matrix element
 */

void QDP_F3_P_eq_elem_C( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F_Complex *src, int ic, int is, int jc, int js, QDP_Subset subset );

/*
 *  Extracting a color vector from a color matrix column
 */

void QDP_F3_V_eq_colorvec_M( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorMatrix *src, int j, QDP_Subset subset );

/*
 *  Inserting a color vector into a color matrix column
 */

void QDP_F3_M_eq_colorvec_V( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorVector *src, int j, QDP_Subset subset );

/*
 *  Extracting a color vector from a half fermion or Dirac fermion
 */

void QDP_F3_V_eq_colorvec_H( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_HalfFermion *src, int spin, QDP_Subset subset );
void QDP_F3_V_eq_colorvec_D( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_DiracFermion *src, int spin, QDP_Subset subset );

/*
 *  Inserting a color vector into a half fermion or Dirac fermion
 */

void QDP_F3_H_eq_colorvec_V( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_ColorVector *src, int spin, QDP_Subset subset );
void QDP_F3_D_eq_colorvec_V( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorVector *src, int spin, QDP_Subset subset );

/*
 *  Extracting a Dirac vector from a Dirac propagator matrix column
 */

void QDP_F3_D_eq_diracvec_P( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracPropagator *src, int jc, int js, QDP_Subset subset );

/*
 *  Inserting a Dirac vector into a Dirac propagator matrix column
 */

void QDP_F3_P_eq_diracvec_D( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracFermion *src, int jc, int js, QDP_Subset subset );

/*
 *  Trace of color matrix
 */

void QDP_F3_C_eq_trace_M( QDP_F_Complex *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );

/*
 *  Real/Imaginary part of trace of color matrix
 */

void QDP_F3_R_eq_re_trace_M( QDP_F_Real *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_R_eq_im_trace_M( QDP_F_Real *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );

/*
 *  Traceless antihermitian part of color matrix
 */

void QDP_F3_M_eq_antiherm_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );

/*
 *  Spin trace of Dirac propagator
 */

void QDP_F3_M_eq_spintrace_P( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );

/*
 *  Dirac spin projection
 */

void QDP_F3_H_eq_spproj_D( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_DiracFermion *src, int dir, int sign, QDP_Subset subset );
void QDP_F3_H_veq_spproj_D( QDP_F3_HalfFermion *__restrict__ dest[], QDP_F3_DiracFermion *src[], int dir[], int sign[], QDP_Subset subset, int nv );

/*
 *  Dirac spin reconstruction
 */

void QDP_F3_D_eq_sprecon_H( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_HalfFermion *src, int dir, int sign, QDP_Subset subset );
void QDP_F3_D_peq_sprecon_H( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_HalfFermion *src, int dir, int sign, QDP_Subset subset );
void QDP_F3_D_meq_sprecon_H( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_HalfFermion *src, int dir, int sign, QDP_Subset subset );
void QDP_F3_D_eqm_sprecon_H( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_HalfFermion *src, int dir, int sign, QDP_Subset subset );
void QDP_F3_D_veq_sprecon_H( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_HalfFermion *src[], int dir[], int sign[], QDP_Subset subset, int nv );
void QDP_F3_D_vpeq_sprecon_H( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_HalfFermion *src[], int dir[], int sign[], QDP_Subset subset, int nv );
void QDP_F3_D_vmeq_sprecon_H( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_HalfFermion *src[], int dir[], int sign[], QDP_Subset subset, int nv );
void QDP_F3_D_veqm_sprecon_H( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_HalfFermion *src[], int dir[], int sign[], QDP_Subset subset, int nv );

/*
 *  Matrix multiply and Dirac spin reconstruction
 */

void QDP_F3_D_eq_sprecon_M_times_H( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, int dir, int sign, QDP_Subset subset );
void QDP_F3_D_peq_sprecon_M_times_H( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, int dir, int sign, QDP_Subset subset );
void QDP_F3_D_meq_sprecon_M_times_H( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, int dir, int sign, QDP_Subset subset );
void QDP_F3_D_eqm_sprecon_M_times_H( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, int dir, int sign, QDP_Subset subset );
void QDP_F3_D_veq_sprecon_M_times_H( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_HalfFermion *src2[], int dir[], int sign[], QDP_Subset subset, int nv );
void QDP_F3_D_vpeq_sprecon_M_times_H( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_HalfFermion *src2[], int dir[], int sign[], QDP_Subset subset, int nv );
void QDP_F3_D_vmeq_sprecon_M_times_H( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_HalfFermion *src2[], int dir[], int sign[], QDP_Subset subset, int nv );
void QDP_F3_D_veqm_sprecon_M_times_H( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_HalfFermion *src2[], int dir[], int sign[], QDP_Subset subset, int nv );

/*
 *  Matrix multiply and Wilson spin multiplication
 */

void QDP_F3_D_eq_wilsonspin_M_times_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, int dir, int sign, QDP_Subset subset );
void QDP_F3_D_peq_wilsonspin_M_times_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, int dir, int sign, QDP_Subset subset );
void QDP_F3_D_meq_wilsonspin_M_times_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, int dir, int sign, QDP_Subset subset );
void QDP_F3_D_eqm_wilsonspin_M_times_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, int dir, int sign, QDP_Subset subset );
void QDP_F3_D_veq_wilsonspin_M_times_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracFermion *src2[], int dir[], int sign[], QDP_Subset subset, int nv );
void QDP_F3_D_vpeq_wilsonspin_M_times_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracFermion *src2[], int dir[], int sign[], QDP_Subset subset, int nv );
void QDP_F3_D_vmeq_wilsonspin_M_times_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracFermion *src2[], int dir[], int sign[], QDP_Subset subset, int nv );
void QDP_F3_D_veqm_wilsonspin_M_times_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracFermion *src2[], int dir[], int sign[], QDP_Subset subset, int nv );

/*
 *  Binary operations with constants
 */

/*
 *  Multiplication by real constant
 */

void QDP_F3_V_eq_r_times_V( QDP_F3_ColorVector *__restrict__ dest, QLA_F_Real *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_eq_r_times_H( QDP_F3_HalfFermion *__restrict__ dest, QLA_F_Real *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_eq_r_times_D( QDP_F3_DiracFermion *__restrict__ dest, QLA_F_Real *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_eq_r_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QLA_F_Real *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eq_r_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QLA_F_Real *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_peq_r_times_V( QDP_F3_ColorVector *__restrict__ dest, QLA_F_Real *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_peq_r_times_H( QDP_F3_HalfFermion *__restrict__ dest, QLA_F_Real *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_peq_r_times_D( QDP_F3_DiracFermion *__restrict__ dest, QLA_F_Real *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_peq_r_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QLA_F_Real *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_peq_r_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QLA_F_Real *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_meq_r_times_V( QDP_F3_ColorVector *__restrict__ dest, QLA_F_Real *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_meq_r_times_H( QDP_F3_HalfFermion *__restrict__ dest, QLA_F_Real *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_meq_r_times_D( QDP_F3_DiracFermion *__restrict__ dest, QLA_F_Real *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_meq_r_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QLA_F_Real *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_meq_r_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QLA_F_Real *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_eqm_r_times_V( QDP_F3_ColorVector *__restrict__ dest, QLA_F_Real *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_eqm_r_times_H( QDP_F3_HalfFermion *__restrict__ dest, QLA_F_Real *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_eqm_r_times_D( QDP_F3_DiracFermion *__restrict__ dest, QLA_F_Real *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_eqm_r_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QLA_F_Real *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eqm_r_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QLA_F_Real *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );

/*
 *  Multiplication by complex constant
 */

void QDP_F3_V_eq_c_times_V( QDP_F3_ColorVector *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_eq_c_times_H( QDP_F3_HalfFermion *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_eq_c_times_D( QDP_F3_DiracFermion *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_eq_c_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eq_c_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_peq_c_times_V( QDP_F3_ColorVector *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_peq_c_times_H( QDP_F3_HalfFermion *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_peq_c_times_D( QDP_F3_DiracFermion *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_peq_c_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_peq_c_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_meq_c_times_V( QDP_F3_ColorVector *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_meq_c_times_H( QDP_F3_HalfFermion *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_meq_c_times_D( QDP_F3_DiracFermion *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_meq_c_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_meq_c_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_eqm_c_times_V( QDP_F3_ColorVector *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_eqm_c_times_H( QDP_F3_HalfFermion *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_eqm_c_times_D( QDP_F3_DiracFermion *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_eqm_c_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eqm_c_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QLA_F_Complex *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );

/*
 *  Multiplication by i
 */

void QDP_F3_V_eq_i_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_H_eq_i_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_D_eq_i_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_M_eq_i_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_eq_i_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_V_peq_i_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_H_peq_i_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_D_peq_i_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_M_peq_i_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_peq_i_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_V_meq_i_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_H_meq_i_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_D_meq_i_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_M_meq_i_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_meq_i_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_V_eqm_i_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_H_eqm_i_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_D_eqm_i_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_M_eqm_i_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_P_eqm_i_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );

/*
 *  Left multiplication by gamma matrix
 */

void QDP_F3_D_eq_gamma_times_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, int i, QDP_Subset subset );
void QDP_F3_P_eq_gamma_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, int i, QDP_Subset subset );

/*
 *  Right multiplication by gamma matrix
 */

void QDP_F3_P_eq_P_times_gamma( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, int i, QDP_Subset subset );

/*
 *  Binary operations with fields
 */

/*
 *  Addition
 */

void QDP_F3_V_eq_V_plus_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_eq_H_plus_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_eq_D_plus_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_eq_M_plus_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eq_P_plus_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );

/*
 *  Subtraction
 */

void QDP_F3_V_eq_V_minus_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_eq_H_minus_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_eq_D_minus_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_eq_M_minus_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eq_P_minus_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );

/*
 *  Multiplication: uniform types
 */

void QDP_F3_P_eq_P_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_P_peq_P_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_P_meq_P_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_P_eqm_P_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );

/*
 *  Left multiplication by color matrix
 */

void QDP_F3_V_eq_M_times_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_eq_M_times_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_eq_M_times_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_eq_M_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eq_M_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_peq_M_times_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_peq_M_times_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_peq_M_times_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_peq_M_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_peq_M_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_meq_M_times_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_meq_M_times_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_meq_M_times_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_meq_M_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_meq_M_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_eqm_M_times_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_eqm_M_times_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_eqm_M_times_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_eqm_M_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eqm_M_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_veq_M_times_V( QDP_F3_ColorVector *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorVector *src2[], QDP_Subset subset, int nv );
void QDP_F3_H_veq_M_times_H( QDP_F3_HalfFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_HalfFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_D_veq_M_times_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_M_veq_M_times_M( QDP_F3_ColorMatrix *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorMatrix *src2[], QDP_Subset subset, int nv );
void QDP_F3_P_veq_M_times_P( QDP_F3_DiracPropagator *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracPropagator *src2[], QDP_Subset subset, int nv );
void QDP_F3_V_vpeq_M_times_V( QDP_F3_ColorVector *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorVector *src2[], QDP_Subset subset, int nv );
void QDP_F3_H_vpeq_M_times_H( QDP_F3_HalfFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_HalfFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_D_vpeq_M_times_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_M_vpeq_M_times_M( QDP_F3_ColorMatrix *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorMatrix *src2[], QDP_Subset subset, int nv );
void QDP_F3_P_vpeq_M_times_P( QDP_F3_DiracPropagator *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracPropagator *src2[], QDP_Subset subset, int nv );
void QDP_F3_V_vmeq_M_times_V( QDP_F3_ColorVector *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorVector *src2[], QDP_Subset subset, int nv );
void QDP_F3_H_vmeq_M_times_H( QDP_F3_HalfFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_HalfFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_D_vmeq_M_times_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_M_vmeq_M_times_M( QDP_F3_ColorMatrix *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorMatrix *src2[], QDP_Subset subset, int nv );
void QDP_F3_P_vmeq_M_times_P( QDP_F3_DiracPropagator *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracPropagator *src2[], QDP_Subset subset, int nv );
void QDP_F3_V_veqm_M_times_V( QDP_F3_ColorVector *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorVector *src2[], QDP_Subset subset, int nv );
void QDP_F3_H_veqm_M_times_H( QDP_F3_HalfFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_HalfFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_D_veqm_M_times_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_M_veqm_M_times_M( QDP_F3_ColorMatrix *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorMatrix *src2[], QDP_Subset subset, int nv );
void QDP_F3_P_veqm_M_times_P( QDP_F3_DiracPropagator *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracPropagator *src2[], QDP_Subset subset, int nv );

/*
 *  Left multiplication by adjoint of color matrix
 */

void QDP_F3_V_eq_Ma_times_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_eq_Ma_times_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_eq_Ma_times_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_eq_Ma_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eq_Ma_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_peq_Ma_times_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_peq_Ma_times_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_peq_Ma_times_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_peq_Ma_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_peq_Ma_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_meq_Ma_times_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_meq_Ma_times_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_meq_Ma_times_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_meq_Ma_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_meq_Ma_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_eqm_Ma_times_V( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_eqm_Ma_times_H( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_eqm_Ma_times_D( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_eqm_Ma_times_M( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eqm_Ma_times_P( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_veq_Ma_times_V( QDP_F3_ColorVector *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorVector *src2[], QDP_Subset subset, int nv );
void QDP_F3_H_veq_Ma_times_H( QDP_F3_HalfFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_HalfFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_D_veq_Ma_times_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_M_veq_Ma_times_M( QDP_F3_ColorMatrix *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorMatrix *src2[], QDP_Subset subset, int nv );
void QDP_F3_P_veq_Ma_times_P( QDP_F3_DiracPropagator *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracPropagator *src2[], QDP_Subset subset, int nv );
void QDP_F3_V_vpeq_Ma_times_V( QDP_F3_ColorVector *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorVector *src2[], QDP_Subset subset, int nv );
void QDP_F3_H_vpeq_Ma_times_H( QDP_F3_HalfFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_HalfFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_D_vpeq_Ma_times_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_M_vpeq_Ma_times_M( QDP_F3_ColorMatrix *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorMatrix *src2[], QDP_Subset subset, int nv );
void QDP_F3_P_vpeq_Ma_times_P( QDP_F3_DiracPropagator *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracPropagator *src2[], QDP_Subset subset, int nv );
void QDP_F3_V_vmeq_Ma_times_V( QDP_F3_ColorVector *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorVector *src2[], QDP_Subset subset, int nv );
void QDP_F3_H_vmeq_Ma_times_H( QDP_F3_HalfFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_HalfFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_D_vmeq_Ma_times_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_M_vmeq_Ma_times_M( QDP_F3_ColorMatrix *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorMatrix *src2[], QDP_Subset subset, int nv );
void QDP_F3_P_vmeq_Ma_times_P( QDP_F3_DiracPropagator *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracPropagator *src2[], QDP_Subset subset, int nv );
void QDP_F3_V_veqm_Ma_times_V( QDP_F3_ColorVector *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorVector *src2[], QDP_Subset subset, int nv );
void QDP_F3_H_veqm_Ma_times_H( QDP_F3_HalfFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_HalfFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_D_veqm_Ma_times_D( QDP_F3_DiracFermion *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_M_veqm_Ma_times_M( QDP_F3_ColorMatrix *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorMatrix *src2[], QDP_Subset subset, int nv );
void QDP_F3_P_veqm_Ma_times_P( QDP_F3_DiracPropagator *__restrict__ dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_DiracPropagator *src2[], QDP_Subset subset, int nv );

/*
 *  Right multiplication by color matrix
 */

void QDP_F3_P_eq_P_times_M( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_peq_P_times_M( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_meq_P_times_M( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eqm_P_times_M( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );

/*
 *  Right multiplication by adjoint of color matrix
 */

void QDP_F3_M_eq_M_times_Ma( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eq_P_times_Ma( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_M_peq_M_times_Ma( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_peq_P_times_Ma( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_M_meq_M_times_Ma( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_meq_P_times_Ma( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_M_eqm_M_times_Ma( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eqm_P_times_Ma( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );

/*
 *  Adjoint of color matrix times adjoint of color matrix
 */

void QDP_F3_M_eq_Ma_times_Ma( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_M_peq_Ma_times_Ma( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_M_meq_Ma_times_Ma( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_M_eqm_Ma_times_Ma( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );

/*
 *  Local inner product
 */

void QDP_F3_C_eq_V_dot_V( QDP_F_Complex *__restrict__ dest, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_C_eq_H_dot_H( QDP_F_Complex *__restrict__ dest, QDP_F3_HalfFermion *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_C_eq_D_dot_D( QDP_F_Complex *__restrict__ dest, QDP_F3_DiracFermion *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_C_eq_M_dot_M( QDP_F_Complex *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_C_eq_P_dot_P( QDP_F_Complex *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );

/*
 *  Real part of local inner product
 */

void QDP_F3_R_eq_re_V_dot_V( QDP_F_Real *__restrict__ dest, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_R_eq_re_H_dot_H( QDP_F_Real *__restrict__ dest, QDP_F3_HalfFermion *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_R_eq_re_D_dot_D( QDP_F_Real *__restrict__ dest, QDP_F3_DiracFermion *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_R_eq_re_M_dot_M( QDP_F_Real *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_R_eq_re_P_dot_P( QDP_F_Real *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );

/*
 *  Color matrix from outer product
 */

void QDP_F3_M_eq_V_times_Va( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_M_peq_V_times_Va( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_M_meq_V_times_Va( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_M_eqm_V_times_Va( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );

/*
 *  Ternary operations with fields
 */

/*
 *  Addition or subtraction with real scalar multiplication
 */

void QDP_F3_V_eq_r_times_V_plus_V( QDP_F3_ColorVector *__restrict__ dest, QLA_F_Real *c, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_eq_r_times_H_plus_H( QDP_F3_HalfFermion *__restrict__ dest, QLA_F_Real *c, QDP_F3_HalfFermion *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_eq_r_times_D_plus_D( QDP_F3_DiracFermion *__restrict__ dest, QLA_F_Real *c, QDP_F3_DiracFermion *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_eq_r_times_M_plus_M( QDP_F3_ColorMatrix *__restrict__ dest, QLA_F_Real *c, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eq_r_times_P_plus_P( QDP_F3_DiracPropagator *__restrict__ dest, QLA_F_Real *c, QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_eq_r_times_V_minus_V( QDP_F3_ColorVector *__restrict__ dest, QLA_F_Real *c, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_eq_r_times_H_minus_H( QDP_F3_HalfFermion *__restrict__ dest, QLA_F_Real *c, QDP_F3_HalfFermion *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_eq_r_times_D_minus_D( QDP_F3_DiracFermion *__restrict__ dest, QLA_F_Real *c, QDP_F3_DiracFermion *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_eq_r_times_M_minus_M( QDP_F3_ColorMatrix *__restrict__ dest, QLA_F_Real *c, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eq_r_times_P_minus_P( QDP_F3_DiracPropagator *__restrict__ dest, QLA_F_Real *c, QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );

/*
 *  Addition or subtraction with complex scalar multiplication
 */

void QDP_F3_V_eq_c_times_V_plus_V( QDP_F3_ColorVector *__restrict__ dest, QLA_F_Complex *c, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_eq_c_times_H_plus_H( QDP_F3_HalfFermion *__restrict__ dest, QLA_F_Complex *c, QDP_F3_HalfFermion *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_eq_c_times_D_plus_D( QDP_F3_DiracFermion *__restrict__ dest, QLA_F_Complex *c, QDP_F3_DiracFermion *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_eq_c_times_M_plus_M( QDP_F3_ColorMatrix *__restrict__ dest, QLA_F_Complex *c, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eq_c_times_P_plus_P( QDP_F3_DiracPropagator *__restrict__ dest, QLA_F_Complex *c, QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_V_eq_c_times_V_minus_V( QDP_F3_ColorVector *__restrict__ dest, QLA_F_Complex *c, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_H_eq_c_times_H_minus_H( QDP_F3_HalfFermion *__restrict__ dest, QLA_F_Complex *c, QDP_F3_HalfFermion *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_D_eq_c_times_D_minus_D( QDP_F3_DiracFermion *__restrict__ dest, QLA_F_Complex *c, QDP_F3_DiracFermion *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_M_eq_c_times_M_minus_M( QDP_F3_ColorMatrix *__restrict__ dest, QLA_F_Complex *c, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_P_eq_c_times_P_minus_P( QDP_F3_DiracPropagator *__restrict__ dest, QLA_F_Complex *c, QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );

/*
 *  Boolean operations
 */

/*
 *  Copy under a mask
 */

void QDP_F3_V_eq_V_mask_I( QDP_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_F3_H_eq_H_mask_I( QDP_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_F3_D_eq_D_mask_I( QDP_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_F3_M_eq_M_mask_I( QDP_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_F3_P_eq_P_mask_I( QDP_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_Int *src2, QDP_Subset subset );

/*
 *  Reductions
 */

/*
 *  Global squared norm: uniform precision
 */

void QDP_F3_r_eq_norm2_V( QLA_F_Real *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_r_eq_norm2_H( QLA_F_Real *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_r_eq_norm2_D( QLA_F_Real *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_r_eq_norm2_M( QLA_F_Real *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_r_eq_norm2_P( QLA_F_Real *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );
void QDP_F3_r_veq_norm2_V( QLA_F_Real dest[], QDP_F3_ColorVector *src[], QDP_Subset subset, int nv );
void QDP_F3_r_veq_norm2_H( QLA_F_Real dest[], QDP_F3_HalfFermion *src[], QDP_Subset subset, int nv );
void QDP_F3_r_veq_norm2_D( QLA_F_Real dest[], QDP_F3_DiracFermion *src[], QDP_Subset subset, int nv );
void QDP_F3_r_veq_norm2_M( QLA_F_Real dest[], QDP_F3_ColorMatrix *src[], QDP_Subset subset, int nv );
void QDP_F3_r_veq_norm2_P( QLA_F_Real dest[], QDP_F3_DiracPropagator *src[], QDP_Subset subset, int nv );

/*
 *  Global inner product
 */

void QDP_F3_c_eq_V_dot_V( QLA_F_Complex *__restrict__ dest, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_c_eq_H_dot_H( QLA_F_Complex *__restrict__ dest, QDP_F3_HalfFermion *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_c_eq_D_dot_D( QLA_F_Complex *__restrict__ dest, QDP_F3_DiracFermion *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_c_eq_M_dot_M( QLA_F_Complex *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_c_eq_P_dot_P( QLA_F_Complex *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_c_veq_V_dot_V( QLA_F_Complex dest[], QDP_F3_ColorVector *src1[], QDP_F3_ColorVector *src2[], QDP_Subset subset, int nv );
void QDP_F3_c_veq_H_dot_H( QLA_F_Complex dest[], QDP_F3_HalfFermion *src1[], QDP_F3_HalfFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_c_veq_D_dot_D( QLA_F_Complex dest[], QDP_F3_DiracFermion *src1[], QDP_F3_DiracFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_c_veq_M_dot_M( QLA_F_Complex dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorMatrix *src2[], QDP_Subset subset, int nv );
void QDP_F3_c_veq_P_dot_P( QLA_F_Complex dest[], QDP_F3_DiracPropagator *src1[], QDP_F3_DiracPropagator *src2[], QDP_Subset subset, int nv );

/*
 *  Real part of global inner product
 */

void QDP_F3_r_eq_re_V_dot_V( QLA_F_Real *__restrict__ dest, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset );
void QDP_F3_r_eq_re_H_dot_H( QLA_F_Real *__restrict__ dest, QDP_F3_HalfFermion *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset );
void QDP_F3_r_eq_re_D_dot_D( QLA_F_Real *__restrict__ dest, QDP_F3_DiracFermion *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset );
void QDP_F3_r_eq_re_M_dot_M( QLA_F_Real *__restrict__ dest, QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset );
void QDP_F3_r_eq_re_P_dot_P( QLA_F_Real *__restrict__ dest, QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset );
void QDP_F3_r_veq_re_V_dot_V( QLA_F_Real dest[], QDP_F3_ColorVector *src1[], QDP_F3_ColorVector *src2[], QDP_Subset subset, int nv );
void QDP_F3_r_veq_re_H_dot_H( QLA_F_Real dest[], QDP_F3_HalfFermion *src1[], QDP_F3_HalfFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_r_veq_re_D_dot_D( QLA_F_Real dest[], QDP_F3_DiracFermion *src1[], QDP_F3_DiracFermion *src2[], QDP_Subset subset, int nv );
void QDP_F3_r_veq_re_M_dot_M( QLA_F_Real dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorMatrix *src2[], QDP_Subset subset, int nv );
void QDP_F3_r_veq_re_P_dot_P( QLA_F_Real dest[], QDP_F3_DiracPropagator *src1[], QDP_F3_DiracPropagator *src2[], QDP_Subset subset, int nv );

/*
 *  Global sums
 */

void QDP_F3_v_eq_sum_V( QLA_F3_ColorVector *__restrict__ dest, QDP_F3_ColorVector *src, QDP_Subset subset );
void QDP_F3_h_eq_sum_H( QLA_F3_HalfFermion *__restrict__ dest, QDP_F3_HalfFermion *src, QDP_Subset subset );
void QDP_F3_d_eq_sum_D( QLA_F3_DiracFermion *__restrict__ dest, QDP_F3_DiracFermion *src, QDP_Subset subset );
void QDP_F3_m_eq_sum_M( QLA_F3_ColorMatrix *__restrict__ dest, QDP_F3_ColorMatrix *src, QDP_Subset subset );
void QDP_F3_p_eq_sum_P( QLA_F3_DiracPropagator *__restrict__ dest, QDP_F3_DiracPropagator *src, QDP_Subset subset );

/*
 *  Multisubset Norms
 */

void QDP_F3_r_eq_norm2_V_multi( QLA_F_Real dest[], QDP_F3_ColorVector *src, QDP_Subset subset[], int ns );
void QDP_F3_r_eq_norm2_H_multi( QLA_F_Real dest[], QDP_F3_HalfFermion *src, QDP_Subset subset[], int ns );
void QDP_F3_r_eq_norm2_D_multi( QLA_F_Real dest[], QDP_F3_DiracFermion *src, QDP_Subset subset[], int ns );
void QDP_F3_r_eq_norm2_M_multi( QLA_F_Real dest[], QDP_F3_ColorMatrix *src, QDP_Subset subset[], int ns );
void QDP_F3_r_eq_norm2_P_multi( QLA_F_Real dest[], QDP_F3_DiracPropagator *src, QDP_Subset subset[], int ns );
void QDP_F3_r_veq_norm2_V_multi( QLA_F_Real dest[], QDP_F3_ColorVector *src[], QDP_Subset subset[], int ns );
void QDP_F3_r_veq_norm2_H_multi( QLA_F_Real dest[], QDP_F3_HalfFermion *src[], QDP_Subset subset[], int ns );
void QDP_F3_r_veq_norm2_D_multi( QLA_F_Real dest[], QDP_F3_DiracFermion *src[], QDP_Subset subset[], int ns );
void QDP_F3_r_veq_norm2_M_multi( QLA_F_Real dest[], QDP_F3_ColorMatrix *src[], QDP_Subset subset[], int ns );
void QDP_F3_r_veq_norm2_P_multi( QLA_F_Real dest[], QDP_F3_DiracPropagator *src[], QDP_Subset subset[], int ns );

/*
 *  Multisubset inner products
 */

void QDP_F3_c_eq_V_dot_V_multi( QLA_F_Complex dest[], QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset[], int ns );
void QDP_F3_c_eq_H_dot_H_multi( QLA_F_Complex dest[], QDP_F3_HalfFermion *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset[], int ns );
void QDP_F3_c_eq_D_dot_D_multi( QLA_F_Complex dest[], QDP_F3_DiracFermion *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset[], int ns );
void QDP_F3_c_eq_M_dot_M_multi( QLA_F_Complex dest[], QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset[], int ns );
void QDP_F3_c_eq_P_dot_P_multi( QLA_F_Complex dest[], QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset[], int ns );
void QDP_F3_c_veq_V_dot_V_multi( QLA_F_Complex dest[], QDP_F3_ColorVector *src1[], QDP_F3_ColorVector *src2[], QDP_Subset subset[], int ns );
void QDP_F3_c_veq_H_dot_H_multi( QLA_F_Complex dest[], QDP_F3_HalfFermion *src1[], QDP_F3_HalfFermion *src2[], QDP_Subset subset[], int ns );
void QDP_F3_c_veq_D_dot_D_multi( QLA_F_Complex dest[], QDP_F3_DiracFermion *src1[], QDP_F3_DiracFermion *src2[], QDP_Subset subset[], int ns );
void QDP_F3_c_veq_M_dot_M_multi( QLA_F_Complex dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorMatrix *src2[], QDP_Subset subset[], int ns );
void QDP_F3_c_veq_P_dot_P_multi( QLA_F_Complex dest[], QDP_F3_DiracPropagator *src1[], QDP_F3_DiracPropagator *src2[], QDP_Subset subset[], int ns );

/*
 *  Multisubset real part of global inner product
 */

void QDP_F3_r_eq_re_V_dot_V_multi( QLA_F_Real dest[], QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset[], int ns );
void QDP_F3_r_eq_re_H_dot_H_multi( QLA_F_Real dest[], QDP_F3_HalfFermion *src1, QDP_F3_HalfFermion *src2, QDP_Subset subset[], int ns );
void QDP_F3_r_eq_re_D_dot_D_multi( QLA_F_Real dest[], QDP_F3_DiracFermion *src1, QDP_F3_DiracFermion *src2, QDP_Subset subset[], int ns );
void QDP_F3_r_eq_re_M_dot_M_multi( QLA_F_Real dest[], QDP_F3_ColorMatrix *src1, QDP_F3_ColorMatrix *src2, QDP_Subset subset[], int ns );
void QDP_F3_r_eq_re_P_dot_P_multi( QLA_F_Real dest[], QDP_F3_DiracPropagator *src1, QDP_F3_DiracPropagator *src2, QDP_Subset subset[], int ns );
void QDP_F3_r_veq_re_V_dot_V_multi( QLA_F_Real dest[], QDP_F3_ColorVector *src1[], QDP_F3_ColorVector *src2[], QDP_Subset subset[], int ns );
void QDP_F3_r_veq_re_H_dot_H_multi( QLA_F_Real dest[], QDP_F3_HalfFermion *src1[], QDP_F3_HalfFermion *src2[], QDP_Subset subset[], int ns );
void QDP_F3_r_veq_re_D_dot_D_multi( QLA_F_Real dest[], QDP_F3_DiracFermion *src1[], QDP_F3_DiracFermion *src2[], QDP_Subset subset[], int ns );
void QDP_F3_r_veq_re_M_dot_M_multi( QLA_F_Real dest[], QDP_F3_ColorMatrix *src1[], QDP_F3_ColorMatrix *src2[], QDP_Subset subset[], int ns );
void QDP_F3_r_veq_re_P_dot_P_multi( QLA_F_Real dest[], QDP_F3_DiracPropagator *src1[], QDP_F3_DiracPropagator *src2[], QDP_Subset subset[], int ns );

/*
 *  Multisubset global sums
 */

void QDP_F3_v_eq_sum_V_multi( QLA_F3_ColorVector dest[], QDP_F3_ColorVector *src, QDP_Subset subset[], int ns );
void QDP_F3_h_eq_sum_H_multi( QLA_F3_HalfFermion dest[], QDP_F3_HalfFermion *src, QDP_Subset subset[], int ns );
void QDP_F3_d_eq_sum_D_multi( QLA_F3_DiracFermion dest[], QDP_F3_DiracFermion *src, QDP_Subset subset[], int ns );
void QDP_F3_m_eq_sum_M_multi( QLA_F3_ColorMatrix dest[], QDP_F3_ColorMatrix *src, QDP_Subset subset[], int ns );
void QDP_F3_p_eq_sum_P_multi( QLA_F3_DiracPropagator dest[], QDP_F3_DiracPropagator *src, QDP_Subset subset[], int ns );

#ifdef __cplusplus
}
#endif

#ifdef QDP_PROFILE
#include "qdp_f3_profile.h"
#endif

#endif
