/**********************

 QDP_INT header file

**********************/

#ifndef _QDP_INT
#define _QDP_INT

#include "qdp_common.h"
#include "qdp_types.h"
#include "qla_types.h"
#include "qla_complex.h"
#include "qla_random.h"

#ifdef __cplusplus
extern "C" {
#endif

/* create and destroy fields */

QDP_RandomState * QDP_create_S(void);
void QDP_destroy_S(QDP_RandomState *field);
QDP_Int * QDP_create_I(void);
void QDP_destroy_I(QDP_Int *field);

/* data access and extraction */

QLA_RandomState * QDP_expose_S(QDP_RandomState *field);
void QDP_reset_S(QDP_RandomState *field);
void QDP_extract_S(QLA_RandomState *dest, QDP_RandomState *src, QDP_Subset subset);
void QDP_insert_S(QDP_RandomState *dest, QLA_RandomState *src, QDP_Subset subset);
QLA_Int * QDP_expose_I(QDP_Int *field);
void QDP_reset_I(QDP_Int *field);
void QDP_extract_I(QLA_Int *dest, QDP_Int *src, QDP_Subset subset);
void QDP_insert_I(QDP_Int *dest, QLA_Int *src, QDP_Subset subset);

/* optimization */

void QDP_discard_I(QDP_Int *field);

/* set by function */

void QDP_I_eq_func(QDP_Int *dest, void (*func)(QLA_Int *dest, int coords[]), QDP_Subset subset );

/* shift */

void QDP_I_eq_sI(QDP_Int *__restrict__ dest, QDP_Int *src, QDP_Shift shift, QDP_ShiftDir fb, QDP_Subset subset);
void QDP_I_veq_sI(QDP_Int *__restrict__ dest[], QDP_Int *src[], QDP_Shift shift[], QDP_ShiftDir fb[], QDP_Subset subset, int nv);

/* shift & multiply */


/*
 *  Fills and random numbers
 */

/*
 *  Zero fills
 */

void QDP_I_eq_zero( QDP_Int *__restrict__ dest, QDP_Subset subset );

/*
 *  Constant fills
 */

void QDP_I_eq_i( QDP_Int *__restrict__ dest, QLA_Int *src, QDP_Subset subset );

/*
 *  Seeding the random number generator field from an integer field
 */

void QDP_S_eq_seed_i_I( QDP_RandomState *__restrict__ dest, QLA_Int c, QDP_Int *src, QDP_Subset subset );

/*
 *  Unary Operations
 */

/*
 *  Bitwise not
 */

void QDP_I_eq_not_I( QDP_Int *__restrict__ dest, QDP_Int *src, QDP_Subset subset );

/*
 *  Copying
 */

void QDP_S_eq_S( QDP_RandomState *__restrict__ dest, QDP_RandomState *src, QDP_Subset subset );
void QDP_I_eq_I( QDP_Int *__restrict__ dest, QDP_Int *src, QDP_Subset subset );
void QDP_S_veq_S( QDP_RandomState *__restrict__ dest[], QDP_RandomState *src[], QDP_Subset subset, int nv );
void QDP_I_veq_I( QDP_Int *__restrict__ dest[], QDP_Int *src[], QDP_Subset subset, int nv );

/*
 *  Incrementing
 */

void QDP_I_eqm_I( QDP_Int *__restrict__ dest, QDP_Int *src, QDP_Subset subset );
void QDP_I_peq_I( QDP_Int *__restrict__ dest, QDP_Int *src, QDP_Subset subset );
void QDP_I_meq_I( QDP_Int *__restrict__ dest, QDP_Int *src, QDP_Subset subset );
void QDP_I_veqm_I( QDP_Int *__restrict__ dest[], QDP_Int *src[], QDP_Subset subset, int nv );
void QDP_I_vpeq_I( QDP_Int *__restrict__ dest[], QDP_Int *src[], QDP_Subset subset, int nv );
void QDP_I_vmeq_I( QDP_Int *__restrict__ dest[], QDP_Int *src[], QDP_Subset subset, int nv );

/*
 *  Binary operations with constants
 */

/*
 *  Multiplication by integer constant
 */

void QDP_I_eq_i_times_I( QDP_Int *__restrict__ dest, QLA_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_peq_i_times_I( QDP_Int *__restrict__ dest, QLA_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_meq_i_times_I( QDP_Int *__restrict__ dest, QLA_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eqm_i_times_I( QDP_Int *__restrict__ dest, QLA_Int *src1, QDP_Int *src2, QDP_Subset subset );

/*
 *  Binary operations with fields
 */

/*
 *  Elementary binary functions on integers
 */

void QDP_I_eq_I_lshift_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eq_I_rshift_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eq_I_mod_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eq_I_max_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eq_I_min_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eq_I_or_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eq_I_and_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eq_I_xor_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );

/*
 *  Addition
 */

void QDP_I_eq_I_plus_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );

/*
 *  Subtraction
 */

void QDP_I_eq_I_minus_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );

/*
 *  Multiplication: uniform types
 */

void QDP_I_eq_I_times_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_peq_I_times_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_meq_I_times_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eqm_I_times_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );

/*
 *  Division of integer, real, and complex fields
 */

void QDP_I_eq_I_divide_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );

/*
 *  Ternary operations with fields
 */

/*
 *  Addition or subtraction with integer scalar multiplication
 */

void QDP_I_eq_i_times_I_plus_I( QDP_Int *__restrict__ dest, QLA_Int *c, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eq_i_times_I_minus_I( QDP_Int *__restrict__ dest, QLA_Int *c, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );

/*
 *  Boolean operations
 */

/*
 *  Comparisons of integers and reals
 */

void QDP_I_eq_I_eq_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eq_I_ne_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eq_I_gt_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eq_I_lt_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eq_I_ge_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );
void QDP_I_eq_I_le_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );

/*
 *  Copy under a mask
 */

void QDP_I_eq_I_mask_I( QDP_Int *__restrict__ dest, QDP_Int *src1, QDP_Int *src2, QDP_Subset subset );

#ifdef __cplusplus
}
#endif

#ifdef QDP_PROFILE
#include "qdp_int_profile.h"
#endif

#endif
