/*
 *  Ternary operations with fields
 *
 *  Addition or subtraction with complex scalar multiplication
 */

#include "qdp_f3_internal.h"

void
QDP_F3_V_eq_c_times_V_plus_V( QDP_F3_ColorVector *__restrict__ dest, QLA_F_Complex *c, QDP_F3_ColorVector *src1, QDP_F3_ColorVector *src2, QDP_Subset subset )
{
  QDP_prepare_dest(&dest->dc);
  QDP_prepare_src(&src1->dc);
  QDP_prepare_src(&src2->dc);

  if( subset->indexed ) {
    if( src1->ptr ) {
      if( src2->ptr ) {
        //QDP_math_time -= QDP_time();
        QLA_F3_V_xeq_c_times_pV_plus_pV( dest->data, c, src1->ptr, src2->ptr, subset->index, subset->len );
        //QDP_math_time += QDP_time();
      } else {
        //QDP_math_time -= QDP_time();
        QLA_F3_V_xeq_c_times_pV_plus_V( dest->data, c, src1->ptr, src2->data, subset->index, subset->len );
        //QDP_math_time += QDP_time();
      }
    } else {
      if( src2->ptr ) {
        //QDP_math_time -= QDP_time();
        QLA_F3_V_xeq_c_times_V_plus_pV( dest->data, c, src1->data, src2->ptr, subset->index, subset->len );
        //QDP_math_time += QDP_time();
      } else {
        //QDP_math_time -= QDP_time();
        QLA_F3_V_xeq_c_times_V_plus_V( dest->data, c, src1->data, src2->data, subset->index, subset->len );
        //QDP_math_time += QDP_time();
      }
    }
  } else {
    if( src1->ptr ) {
      if( src2->ptr ) {
        //QDP_math_time -= QDP_time();
        QLA_F3_V_veq_c_times_pV_plus_pV( dest->data+subset->offset, c, src1->ptr+subset->offset, src2->ptr+subset->offset, subset->len );
        //QDP_math_time += QDP_time();
      } else {
        //QDP_math_time -= QDP_time();
        QLA_F3_V_veq_c_times_pV_plus_V( dest->data+subset->offset, c, src1->ptr+subset->offset, src2->data+subset->offset, subset->len );
        //QDP_math_time += QDP_time();
      }
    } else {
      if( src2->ptr ) {
        //QDP_math_time -= QDP_time();
        QLA_F3_V_veq_c_times_V_plus_pV( dest->data+subset->offset, c, src1->data+subset->offset, src2->ptr+subset->offset, subset->len );
        //QDP_math_time += QDP_time();
      } else {
        //QDP_math_time -= QDP_time();
        QLA_F3_V_veq_c_times_V_plus_V( dest->data+subset->offset, c, src1->data+subset->offset, src2->data+subset->offset, subset->len );
        //QDP_math_time += QDP_time();
      }
    }
  }
}
